import { mix, Mixin } from '../mixin.js'

/** @import { MixinFunction } from '../mixin.js' */
/** @import { Fn } from './types.js' */

export const ProtectedConstructor = Symbol('ProtectedConstructor')
export class ProtectedConstructorError extends Error {
  /** @param {string} name */
  constructor(name) {
    super(`ProtectedConstructorError: ${name} cannot be directly constructed`)
  }
}

class Derivations {
  #inner

  /**
   * @param {Iterable<readonly [any, any]>} iterable 
   */
  constructor(iterable) {
    this.#inner = new Map(iterable)
  }

  /**
   * @param {Iterable<any>} keys 
   * @param {any} value 
   * @returns {this}
   */
  set(keys, value) {
    this.#inner.set(new Set(keys), value)
    return this
  }

  /**
   * @param {Iterable<any>} keys 
   * @returns {boolean}
   */
  has(keys) {
    return this.#inner.has(new Set(keys))
  }

  /**
   * @param {Iterable<any>} keys 
   */
  get(keys) {
    return this.#inner.get(new Set(keys))
  }

  /**
   * @param {Iterable<any>} keys 
   */
  delete(keys) {
    return this.#inner.delete(new Set(keys))
  }
}

export class NotImplementedError extends Error {
  /** @param {string} name */
  constructor(name) {
    super(`${name} is not implemented`)
  }
}

/** @typedef {(mixin: MixinFunction) => MixinFunction} MixinWrapper */

/**
 * @enum {number}
 */
const MethodType = Object.freeze({
  Instance: 0,
  Static: 1
})

class Method {
  #name

  /** @type {MethodType} */
  #type = MethodType.Instance

  /** @type {Fn} */
  #implementation

  /**
   * @param {string | Method} name 
   */
  constructor(name) {
    if (name instanceof Method) {
      return name
    }

    this.#name = name
  }

  /**
   * @param {string | Method} value 
   */
  static from(value) {
    return new Method(value)
  }

  isInstance() {
    this.#type = MethodType.Instance
    return this
  }

  isStatic() {
    this.#type = MethodType.Static
    return this
  }

  /** @param {Fn} f */
  implementation(f) {
    this.#implementation = f
    return this
  }

  /**
   * @param {string} interfaceName 
   */
  _defaultImplementation(interfaceName) {
    const err = new NotImplementedError(
      `${interfaceName}::${this.#name}`
    )

    return function() { throw err }
  }

  /**
   * @param {Function} target 
   */
  _getInstallationPoint(target) {
    switch (this.#type) {
      case 0: return Object.getPrototypeOf(target)
      case 1: return target
      default: return target
    }
  }

  /**
   * @param {Interface} builder
   * @param {Function} target 
   */
  implement(builder, target) {
    const impl = this.#implementation || this._defaultImplementation(builder.name)

    this._getInstallationPoint(target)[this.#name] = impl
  }
}

const Implementations = Symbol()

class Interface {
  #name

  /** @type {MixinFunction} */
  #mixin

  /** @type {Set<Method>} */
  #methods = new Set()

  /** @type {Set<Interface>} */
  #interfaces = new Set()

  /** @param {string} name */
  constructor(name) {
    this.#name = name
    this.#mixin = Mixin(this.build.bind(this))
  }

  get name() {
    return this.#name
  }

  findInterfaces(type) {
    let interfaces = new Set()
    let current = type

    while (current != null) {
      interfaces = new Set(current[Implementations]).union(interfaces)
      current = Object.getPrototypeOf(current)
    }

    return interfaces
  }

  /**
   * @param {{ name: string }} type 
   * @returns {boolean}
   */
  implementedBy(type) {
    return this.findInterfaces(type).has(this)
  }

  /**
   * @param {...(PropertyKey | Method)} methods
   * @returns {this}
   */
  specifies(...methods) {
    this.#methods = new Set(
      methods
        .map(Method.from)
        .concat(...this.#methods)
    )

    return this
  }

  /**
   * @param {...Interface} interfaces
   * @returns {this}
   */
  extends(...interfaces) {
    this.#interfaces = new Set(interfaces.concat(...this.#interfaces))
    return this
  }

  /**
   * @returns {MixinFunction}
   */
  asMixin() {
    return this.#mixin
  }

  /**
   * @param {FunctionConstructor} base
   */
  build(base) {
    const interfaces = [...this.#interfaces.values()]
    const mixins = interfaces.map(x => x.asMixin())

    const Interfaces = mix(base).with(...mixins)

    const Algebra = class extends Interfaces { }

    for (const method of this.#methods) {
      method.implement(this, Algebra)
    }

    const prototype = Object.getPrototypeOf(Algebra)
    prototype[Implementations] = new Set(interfaces)

    return Algebra
  }
}

export class BaseSet { }

/**
 * @param {PropertyKey} key
 * @param {object} obj
 * @returns {boolean}
 */
export const hasOwn = (key, obj) => Object.hasOwn(obj, key)

/**
 * @param {Function} left
 * @param {Function} right
 */
export const EitherOf = (left, right) => mix(left).with(l => right(l))

/**
 * @param {string} methodName
 * @param {(...args: any[]) => any} f
 * @param {object} obj
 */
export const apply = (methodName, f, obj) => {
  if (hasOwn(methodName, obj)) {
    return obj[methodName](f)
  } else {
    return f(obj)
  }
}

/**
 * @param {Function} base
 * @returns {(...algebras: Interface[]) => FunctionConstructor}
 */
export const AlgebraWithBase = base => (...algebras) => {
  return mix(base).with(...algebras.map(x => x.asMixin()))
}

/**
 * @param {...Interface} algebras
 * @returns {FunctionConstructor}
 */
export const Algebra = AlgebraWithBase(BaseSet)

export const Setoid = new Interface('Setoid')
  .specifies('equals')

export const Ord = new Interface('Ord')
  .specifies('lte')

export const Semigroupoid = new Interface('Semigroupoid')
  .specifies('compose')

export const Category = new Interface('Category')
  .extends(Semigroupoid)
  .specifies(
    Method.from('id').isStatic()
  )

export const Semigroup = new Interface('Semigroup')
  .specifies('concat')

export const Monoid = new Interface('Monoid')
  .extends(Semigroup)
  .specifies(
    Method.from('empty').isStatic()
  )

export const Group = new Interface('Group')
  .extends(Monoid)
  .specifies('invert')

export const Filterable = new Interface('Filterable')
  .specifies('filter')

export const Functor = new Interface('Functor')
  .specifies('map')

export const Contravariant = new Interface('Contravariant')
  .specifies('contramap')

export const Apply = new Interface('Apply')
  .extends(Functor)
  .specifies('ap')

export const Applicative = new Interface('Applicative')
  .extends(Apply)
  .specifies(
    Method.from('of').isStatic()
  )

export const Alt = new Interface('Alt')
  .extends(Functor)
  .specifies('alt')

export const Plus = new Interface('Plus')
  .extends(Alt)
  .specifies(
    Method.from('zero').isStatic()
  )

export const Alternative = new Interface('Alternative')
  .extends(Applicative, Plus)

export const Foldable = new Interface('Foldable')
  .specifies('fold')

export const Traversable = new Interface('Traversable')
  .extends(Functor, Foldable)
  .specifies('traverse')

export const Chain = new Interface('Chain')
  .extends(Apply)
  .specifies(
    Method.from('chain')
      .implementation(function(f) {
        return f(this._value)
      })
  )

export const ChainRef = new Interface('ChainRec')
  .extends(Chain)
  .specifies(
    Method.from('chainRec').isStatic()
  )

export const Monad = new Interface('Monad')
  .extends(Applicative, Chain)

export const Extend = new Interface('Extend')
  .extends(Functor)
  .specifies('extend')

export const Comonad = new Interface('Comonad')
  .extends(Extend)
  .specifies('extract')

export const Bifunctor = new Interface('Bifunctor')
  .extends(Functor)
  .specifies('bimap')

export const Profunctor = new Interface('Profunctor')
  .extends(Functor)
  .specifies('promap')

