import { Algebra, Comonad, Foldable, Monoid, Semigroup } from './interfaces.js'

/** @import { Apply, Chain, Foldable as FoldableT, Functor, Morphism, Semigroup as SemigroupT } from './types.js' */

const Interfaces = Algebra(Semigroup, Foldable, Monoid, Comonad)

/**
 * @template T
 * @typedef {Element<T> | Empty<T>} List
 */

/** @template T */
class Empty extends Interfaces {
  constructor() {
    super()
  }

  /**
   * @template U
   * @type {Chain<T>['chain']}
   * @this {Empty<U>}
   * @param {Morphism<T, List<U>>} _f
   * @returns {List<U>}
   */
  chain(_f) {
    return /** @type {List<U>} */ (this)
  }

  /**
   * @template U
   * @type {Functor<T>['map']}
   * @this {Empty<U>}
   * @param {Morphism<T, U>} _f 
   * @returns {List<U>}
   */
  map(_f) {
    return this
  }

  /**
   * @template U
   * @type {Apply<T>['ap']}
   * @this {Empty<U>}
   * @param {List<Morphism<T, U>>} _b 
   * @returns {List<U>}
   */
  ap(_b) {
    return /** @type {List<U>} */ (this)
  }

  /**
   * @type {SemigroupT<T>['concat']}
   */
  concat(b) {
    return b
  }

  /**
   * @type {FoldableT<T>['reduce']}
   */
  reduce(_f, acc) {
    return acc
  }

  count() {
    return 0
  }

  /** @returns {this is Empty<T>} */
  isEmpty() {
    return true
  }
}

/** @template T */
class Element extends Interfaces {
  #head
  #tail
  /** @type {List<T>} */
  #cache

  /**
   * @param {T} head
   * @param {() => List<T>} [tail]
   */
  constructor(head, tail = List.empty) {
    super()
    this.#head = head
    this.#tail = tail
    this.#cache = null
  }
  /**
   * @template U
   * @type {Chain<T>['chain']}
   * @this {Element<T>}
   * @param {Morphism<T, List<U>>} f 
   * @returns {List<U>}
   */
  chain(f) {
    return /** @type {List<U>} */ (f(this.#head).concat(
      /** @type {never} */(this.tail().chain(f))
    ))
  }

  /**
   * @template U
   * @type {Functor<T>['map']}
   * @param {Morphism<T, U>} f 
   * @returns {List<U>}
   */
  map(f) {
    return new Element(
      f(this.#head),
      () => /** @type {List<U>} */(this.tail().map(f))
    )
  }

  /**
   * @template U
   * @type {Apply<T>['ap']}
   * @this {Element<T>}
   * @param {List<Morphism<T, U>>} b 
   * @returns {List<U>}
   */
  ap(b) {
    if (b.isEmpty()) {
      return List.empty()
    }

    const head = /** @type {List<U>} */ (this.map(b.head()))
    const rest = /** @type {List<U>} */ (this.ap(b.tail()))
    return /** @type {List<U>} */ (head.concat(rest))
  }

  /**
   * @type {SemigroupT<T>['concat']}
   */
  concat(b) {
    return new Element(
      this.#head,
      () => /** @type {List<T>} */(this.tail().concat(b))
    )
  }

  /**
   * @type {FoldableT<T>['reduce']}
   */
  reduce(f, acc) {
    return this.tail().reduce(f, f(acc, this.#head))
  }

  head() {
    return this.#head
  }

  tail() {
    return this.#cache || (this.#cache = this.#tail())
  }

  count() {
    return this.tail().count() + 1
  }

  /** @returns {this is Empty<T>} */
  isEmpty() {
    return false
  }

  toArray() {
    return this.reduce(
      reduceArray,
      []
    )
  }
}

class TypeRef {
  /**
   * @template  T
   * @param {T} value
   * @returns {List<T>}
   */
  static of(value) {
    return new Element(value)
  }

  /**
   * @template T
   * @param {Iterable<T>} iterable
   * @returns {List<T>}
   */
  static from(iterable) {
    const iterator = Iterator.from(iterable)
    return List.fromIterator(iterator)
  }

  /**
   * @template T
   * @param {Iterator<T>} iterator
   * @returns {List<T>}
   */
  static fromIterator(iterator) {
    const next = iterator.next()

    if (next.done) {
      return List.empty()
    } else {
      return new Element(next.value, () => List.fromIterator(iterator))
    }
  }

  /**
   * @template T
   * @param {T} head 
   * @param {List<T>} tail 
   * @returns {List<T>}
   */
  static cons(head, tail) {
    return new Element(head, () => tail)
  }

  static empty() {
    return empty
  }
}

/**
 * @template T
 * @param {T[]} acc
 * @param {T} x
 * @returns {T[]}
 */
const reduceArray = (acc, x) => acc.concat(x)

export const List = TypeRef

const empty = new Empty()

/**
 * @template T
 * @param {T} value
 */
export const list = value => List.of(value)

